module Comparable.Dict exposing
    ( Dict
    , equals
    , empty, singleton, insert, update, remove
    , isEmpty, member, get, size
    , keys, values, toList, fromList
    , map, foldl, foldr, filter, partition
    , union, intersect, diff, merge
    )

{-| A dictionary mapping unique keys to values. The keys can be any type.


# Dictionaries

@docs Dict


# Compare

@docs equals


# Build

@docs empty, singleton, insert, update, remove


# Query

@docs isEmpty, member, get, size


# List

@docs keys, values, toList, fromList


# Transform

@docs map, foldl, foldr, filter, partition


# Combine

@docs union, intersect, diff, merge

-}

import Comparable exposing (Comparable)


type Color
    = Red
    | Black


{-| A dictionary of keys and values.
-}
type Dict key value
    = Node Color key value (Dict key value) (Dict key value)
    | Empty


{-| Determine if two dictionaries contain the same entries.
-}
equals : Dict key value -> Dict key value -> Bool
equals first second =
    toList first == toList second


{-| Create an empty dictionary.
-}
empty : Dict key value
empty =
    Empty


{-| Get the value associated with a key.
-}
get : Comparable key -> key -> Dict key value -> Maybe value
get compare key dict =
    case dict of
        Empty ->
            Nothing

        Node _ nodeKey value left right ->
            case compare key nodeKey of
                LT ->
                    get compare key left

                EQ ->
                    Just value

                GT ->
                    get compare key right


{-| Determine if a key is in a dictionary.
-}
member : Comparable key -> key -> Dict key value -> Bool
member compare key dict =
    case get compare key dict of
        Just _ ->
            True

        Nothing ->
            False


{-| Determine the number of key-value pairs in the dictionary.
-}
size : Dict key value -> Int
size dict =
    sizeHelper 0 dict


sizeHelper : Int -> Dict key value -> Int
sizeHelper n dict =
    case dict of
        Empty ->
            n

        Node _ _ _ left right ->
            sizeHelper (sizeHelper (n + 1) right) left


{-| Determine if a dictionary is empty.
-}
isEmpty : Dict key value -> Bool
isEmpty dict =
    case dict of
        Empty ->
            True

        Node _ _ _ _ _ ->
            False


{-| Insert a key-value pair into a dictionary. Replaces value when there is a collision.
-}
insert : Comparable key -> key -> value -> Dict key value -> Dict key value
insert compare key value dict =
    case insertHelper compare key value dict of
        Node Red nodeKey nodeValue left right ->
            Node Black nodeKey nodeValue left right

        x ->
            x


insertHelper : Comparable key -> key -> value -> Dict key value -> Dict key value
insertHelper compare key value dict =
    case dict of
        Empty ->
            Node Red key value Empty Empty

        Node nodeColor nodeKey nodeValue left right ->
            case compare key nodeKey of
                LT ->
                    balanceHelper nodeColor nodeKey nodeValue (insertHelper compare key value left) right

                EQ ->
                    Node nodeColor nodeKey value left right

                GT ->
                    balanceHelper nodeColor nodeKey nodeValue left (insertHelper compare key value right)


balanceHelper : Color -> key -> value -> Dict key value -> Dict key value -> Dict key value
balanceHelper color key value left right =
    case right of
        Node Red redKey redValue redLeft redRight ->
            case left of
                Node Red leftKey leftValue leftLeft leftRight ->
                    Node
                        Red
                        key
                        value
                        (Node Black leftKey leftValue leftLeft leftRight)
                        (Node Black redKey redValue redLeft redRight)

                _ ->
                    Node color redKey redValue (Node Red key value left redLeft) redRight

        _ ->
            case left of
                Node Red leftKey leftValue (Node Red leftLeftKey leftleftValue leftLeftLeft leftLeftRight) leftRight ->
                    Node
                        Red
                        leftKey
                        leftValue
                        (Node Black leftLeftKey leftleftValue leftLeftLeft leftLeftRight)
                        (Node Black key value leftRight right)

                _ ->
                    Node color key value left right


{-| Remove a key-value pair from a dictionary. If the key is not found, no changes are made.
-}
remove : Comparable key -> key -> Dict key value -> Dict key value
remove compare key dict =
    case removeHelper compare key dict of
        Node Red nodeKey value left right ->
            Node Black nodeKey value left right

        x ->
            x


removeHelper : Comparable key -> key -> Dict key value -> Dict key value
removeHelper compare key dict =
    case dict of
        Empty ->
            Empty

        Node color nodeKey value left right ->
            case compare key nodeKey of
                LT ->
                    case left of
                        Node Black _ _ leftLeft _ ->
                            case leftLeft of
                                Node Red _ _ _ _ ->
                                    Node color nodeKey value (removeHelper compare key left) right

                                _ ->
                                    case moveRedLeftHelper dict of
                                        Node movedColor movedKey movedValue movedLeft movedRight ->
                                            balanceHelper movedColor movedKey movedValue (removeHelper compare key movedLeft) movedRight

                                        Empty ->
                                            Empty

                        _ ->
                            Node color nodeKey value (removeHelper compare key left) right

                _ ->
                    removeHelperEQGT compare key (removeHelperBeforeEQGT dict color nodeKey value left right)


removeHelperBeforeEQGT : Dict key value -> Color -> key -> value -> Dict key value -> Dict key value -> Dict key value
removeHelperBeforeEQGT dict color key value left right =
    case left of
        Node Red leftKey leftValue leftLeft leftRight ->
            Node color leftKey leftValue leftLeft (Node Red key value leftRight right)

        _ ->
            case right of
                Node Black _ _ (Node Black _ _ _ _) _ ->
                    moveRedRightHelper dict

                Node Black _ _ Empty _ ->
                    moveRedRightHelper dict

                _ ->
                    dict


removeHelperEQGT : Comparable key -> key -> Dict key value -> Dict key value
removeHelperEQGT compare key dict =
    case dict of
        Node color nodeKey value left right ->
            case compare key nodeKey of
                EQ ->
                    case getMinHelper right of
                        Node _ minKey minValue _ _ ->
                            balanceHelper color minKey minValue left (removeMinHelper right)

                        Empty ->
                            Empty

                _ ->
                    balanceHelper color nodeKey value left (removeHelper compare key right)

        Empty ->
            Empty


getMinHelper : Dict key value -> Dict key value
getMinHelper dict =
    case dict of
        Node _ _ _ ((Node _ _ _ _ _) as left) _ ->
            getMinHelper left

        _ ->
            dict


removeMinHelper : Dict key value -> Dict key value
removeMinHelper dict =
    case dict of
        Node color key value ((Node leftColor _ _ leftLeft _) as left) right ->
            case leftColor of
                Black ->
                    case leftLeft of
                        Node Red _ _ _ _ ->
                            Node color key value (removeMinHelper left) right

                        _ ->
                            case moveRedLeftHelper dict of
                                Node nodeColor nodeKey nodeValue nodeLeft nodeRight ->
                                    balanceHelper nodeColor nodeKey nodeValue (removeMinHelper nodeLeft) nodeRight

                                Empty ->
                                    Empty

                _ ->
                    Node color key value (removeMinHelper left) right

        _ ->
            Empty


moveRedLeftHelper : Dict key value -> Dict key value
moveRedLeftHelper dict =
    case dict of
        Node _ key value (Node _ leftKey leftValue leftLeft leftRight) (Node _ rightKey rightValue (Node Red rightLeftKey rightLeftValue rightLeftLeft rightLeftRight) rightRight) ->
            Node Red
                rightLeftKey
                rightLeftValue
                (Node Black key value (Node Red leftKey leftValue leftLeft leftRight) rightLeftLeft)
                (Node Black rightKey rightValue rightLeftRight rightRight)

        Node color key value (Node _ leftKey leftValue leftLeft leftRight) (Node _ rightKey rightValue rightLeft rightRight) ->
            case color of
                Black ->
                    Node Black key value (Node Red leftKey leftValue leftLeft leftRight) (Node Red rightKey rightValue rightLeft rightRight)

                Red ->
                    Node Black key value (Node Red leftKey leftValue leftLeft leftRight) (Node Red rightKey rightValue rightLeft rightRight)

        _ ->
            dict


moveRedRightHelper : Dict key value -> Dict key value
moveRedRightHelper dict =
    case dict of
        Node _ key value (Node _ leftKey leftValue (Node Red leftLeftKey leftLeftValue leftLeftLeft leftLeftRight) leftRight) (Node _ rightKey rightValue rightLeft rightRight) ->
            Node Red leftKey leftValue (Node Black leftLeftKey leftLeftValue leftLeftLeft leftLeftRight) (Node Black key value leftRight (Node Red rightKey rightValue rightLeft rightRight))

        Node color key value (Node _ leftKey leftValue leftLeft leftRight) (Node _ rightKey rightValue rightLeft rightRight) ->
            case color of
                Black ->
                    Node Black key value (Node Red leftKey leftValue leftLeft leftRight) (Node Red rightKey rightValue rightLeft rightRight)

                Red ->
                    Node Black key value (Node Red leftKey leftValue leftLeft leftRight) (Node Red rightKey rightValue rightLeft rightRight)

        _ ->
            dict


{-| Update the value of a dictionary for a specific key with a given function.
-}
update : Comparable key -> key -> (Maybe value -> Maybe value) -> Dict key value -> Dict key value
update compare key updateValue dict =
    case updateValue (get compare key dict) of
        Just value ->
            insert compare key value dict

        Nothing ->
            remove compare key dict


{-| Create a dictionary with one key-value pair.
-}
singleton : key -> value -> Dict key value
singleton key value =
    Node Black key value Empty Empty


{-| Combine two dictionaries. If there is a collision, preference is given to the first dictionary.
-}
union : Comparable key -> Dict key value -> Dict key value -> Dict key value
union compare first second =
    foldl (insert compare) second first


{-| Keep a key-value pair when its key appears in the second dictionary. Preference is given to values in the first dictionary.
-}
intersect : Comparable key -> Dict key value -> Dict key value -> Dict key value
intersect compare first second =
    filter compare (\k _ -> member compare k second) first


{-| Keep a key-value pair when its key does not appear in the second dictionary.
-}
diff : Comparable key -> Dict key value -> Dict key otherValue -> Dict key value
diff compare first second =
    foldl (\k _ t -> remove compare k t) first second


{-| The most general way of combining two dictionaries. You provide three accumulators for when a given key appears:

1.  Only in the left dictionary.
2.  In both dictionaries.
3.  Only in the right dictionary.
    You then traverse all the keys from lowest to highest, building up whatever
    you want.

-}
merge :
    Comparable key
    -> (key -> value -> result -> result)
    -> (key -> value -> otherValue -> result -> result)
    -> (key -> otherValue -> result -> result)
    -> Dict key value
    -> Dict key otherValue
    -> result
    -> result
merge compare leftStep bothStep rightStep leftDict rightDict initialResult =
    let
        ( leftovers, intermediateResult ) =
            foldl (mergeHelper compare leftStep bothStep rightStep) ( toList leftDict, initialResult ) rightDict

        fold ( key, value ) result =
            leftStep key value result
    in
    List.foldl fold intermediateResult leftovers


mergeHelper :
    Comparable key
    -> (key -> value -> result -> result)
    -> (key -> value -> otherValue -> result -> result)
    -> (key -> otherValue -> result -> result)
    -> key
    -> otherValue
    -> ( List ( key, value ), result )
    -> ( List ( key, value ), result )
mergeHelper compare leftStep bothStep rightStep rightKey rightValue ( list, result ) =
    case list of
        [] ->
            ( list, rightStep rightKey rightValue result )

        ( leftKey, leftValue ) :: rest ->
            case compare leftKey rightKey of
                LT ->
                    mergeHelper compare leftStep bothStep rightStep rightKey rightValue ( rest, leftStep leftKey leftValue result )

                GT ->
                    ( list, rightStep rightKey rightValue result )

                EQ ->
                    ( rest, bothStep leftKey leftValue rightValue result )


{-| Apply a function to all values in a dictionary.
-}
map : (key -> value -> otherValue) -> Dict key value -> Dict key otherValue
map toOther dict =
    case dict of
        Empty ->
            Empty

        Node color key value left right ->
            Node color key (toOther key value) (map toOther left) (map toOther right)


{-| Fold over the key-value pairs in a dictionary from lowest key to highest key.
-}
foldl : (key -> value -> accumulator -> accumulator) -> accumulator -> Dict key value -> accumulator
foldl toOther accumulator dict =
    case dict of
        Empty ->
            accumulator

        Node _ key value left right ->
            foldl toOther (toOther key value (foldl toOther accumulator left)) right


{-| Fold over the key-value pairs in a dictionary from highest key to lowest key.
-}
foldr : (key -> value -> accumulator -> accumulator) -> accumulator -> Dict key value -> accumulator
foldr toOther accumulator dict =
    case dict of
        Empty ->
            accumulator

        Node _ key value left right ->
            foldr toOther (toOther key value (foldr toOther accumulator right)) left


{-| Keep only the key-value pairs that pass the given test.
-}
filter : Comparable key -> (key -> value -> Bool) -> Dict key value -> Dict key value
filter compare isKept =
    foldl (filterHelper compare isKept) empty


filterHelper : Comparable key -> (key -> value -> Bool) -> key -> value -> Dict key value -> Dict key value
filterHelper compare isKept key value dict =
    if isKept key value then
        insert compare key value dict

    else
        dict


{-| Partition a dictionary according to some test. The first dictionary
contains all key-value pairs which passed the test, and the second contains
the pairs that did not.
-}
partition : Comparable key -> (key -> value -> Bool) -> Dict key value -> ( Dict key value, Dict key value )
partition compare isLeft =
    foldl (partitionHelper compare isLeft) ( empty, empty )


partitionHelper :
    Comparable key
    -> (key -> value -> Bool)
    -> key
    -> value
    -> ( Dict key value, Dict key value )
    -> ( Dict key value, Dict key value )
partitionHelper compare isLeft key value ( left, right ) =
    if isLeft key value then
        ( insert compare key value left, right )

    else
        ( left, insert compare key value right )


{-| Get all of the keys in a dictionary, sorted from lowest to highest.
-}
keys : Dict key value -> List key
keys =
    foldr (\key _ keyList -> key :: keyList) []


{-| Get all of the values in a dictionary, in the compare of their keys.
-}
values : Dict key value -> List value
values =
    foldr (\_ value valueList -> value :: valueList) []


{-| Convert a dictionary into an association list of key-value pairs, sorted by keys.
-}
toList : Dict key value -> List ( key, value )
toList =
    foldr (\key value list -> ( key, value ) :: list) []


{-| Convert an association list into a dictionary.
-}
fromList : Comparable key -> List ( key, value ) -> Dict key value
fromList compare =
    List.foldl (\( key, value ) dict -> insert compare key value dict) empty
